use cosmic_text::{Align, Attrs, CacheKeyFlags, Color, Family, Stretch, Style, Weight};

#[cfg(feature = "databake")]
use databake::Bake;
use jotdown::{ListKind, OrderedListNumbering, OrderedListStyle};
#[cfg(feature = "serde")]
use serde::{Deserialize, Serialize};

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "Option<ListKind>"))]
pub enum ListKindOption {
    Some(#[cfg_attr(feature = "serde", serde(with = "ListKindSerde"))] ListKind),
    None,
}

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "jotdown::ListKind"))]
pub enum ListKindSerde {
    Unordered,
    Ordered {
        numbering: OrderedListNumbering,
        style: OrderedListStyle,
        start: u64,
    },
    Task,
}

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "jotdown::OrderedListNumbering"))]
pub enum OrderedListNumberingSerde {
    Decimal,
    AlphaLower,
    AlphaUpper,
    RomanLower,
    RomanUpper,
}

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "jotdown::OrderedListStyle"))]
pub enum OrderedListStyleSerde {
    Period,
    Paren,
    ParenParen,
}

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "cosmic_text::Metrics"))]
#[cfg_attr(feature = "databake", derive(Bake))]
#[cfg_attr(feature = "databake", databake(path = cosmic_text::Metrics))]
pub struct MetricsSerde {
    pub font_size: f32,
    pub line_height: f32,
}

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "Family"))]
#[cfg_attr(feature = "databake", derive(Bake))]
#[cfg_attr(feature = "databake", databake(path = cosmic_text::Family))]
pub enum FamilySerde<'a> {
    Name(&'a str),
    Serif,
    SansSerif,
    Cursive,
    Fantasy,
    Monospace,
}

impl<'a> From<Family<'a>> for FamilySerde<'a> {
    fn from(value: Family<'a>) -> Self {
        match value {
            Family::Name(n) => FamilySerde::Name(n),
            Family::Serif => FamilySerde::Serif,
            Family::SansSerif => FamilySerde::SansSerif,
            Family::Cursive => FamilySerde::Cursive,
            Family::Fantasy => FamilySerde::Fantasy,
            Family::Monospace => FamilySerde::Monospace,
        }
    }
}

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "Stretch"))]
#[cfg_attr(feature = "databake", derive(Bake))]
#[cfg_attr(feature = "databake", databake(path = cosmic_text::Stretch))]
pub enum StretchSerde {
    UltraCondensed,
    ExtraCondensed,
    Condensed,
    SemiCondensed,
    Normal,
    SemiExpanded,
    Expanded,
    ExtraExpanded,
    UltraExpanded,
}

impl From<Stretch> for StretchSerde {
    fn from(value: Stretch) -> Self {
        match value {
            Stretch::UltraCondensed => StretchSerde::UltraCondensed,
            Stretch::ExtraCondensed => StretchSerde::ExtraCondensed,
            Stretch::Condensed => StretchSerde::Condensed,
            Stretch::SemiCondensed => StretchSerde::SemiCondensed,
            Stretch::Normal => StretchSerde::Normal,
            Stretch::SemiExpanded => StretchSerde::SemiExpanded,
            Stretch::Expanded => StretchSerde::Expanded,
            Stretch::ExtraExpanded => StretchSerde::ExtraExpanded,
            Stretch::UltraExpanded => StretchSerde::UltraExpanded,
        }
    }
}

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "Style"))]
#[cfg_attr(feature = "databake", derive(Bake))]
#[cfg_attr(feature = "databake", databake(path = cosmic_text::Style))]
pub enum StyleSerde {
    Normal,
    Italic,
    Oblique,
}

impl From<Style> for StyleSerde {
    fn from(value: Style) -> Self {
        match value {
            Style::Normal => StyleSerde::Normal,
            Style::Italic => StyleSerde::Italic,
            Style::Oblique => StyleSerde::Oblique,
        }
    }
}

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "Weight"))]
#[cfg_attr(feature = "databake", derive(Bake))]
#[cfg_attr(feature = "databake", databake(path = cosmic_text::Weight))]
pub struct WeightSerde(pub u16);

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "Color"))]
#[cfg_attr(feature = "databake", derive(Bake))]
#[cfg_attr(feature = "databake", databake(path = cosmic_text::Color))]
pub struct ColorSerde(pub u32);

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "Option<Color>"))]
pub enum ColorOpt {
    Some(#[cfg_attr(feature = "serde", serde(with = "ColorSerde"))] Color),
    None,
}

#[cfg(feature = "serde")]
mod cache_key_flags {
    use cosmic_text::CacheKeyFlags;
    use serde::{Deserialize, Deserializer, Serialize, Serializer};

    pub fn serialize<S>(_: &CacheKeyFlags, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        ().serialize(serializer)
    }

    pub fn deserialize<'de, D>(deserializer: D) -> Result<CacheKeyFlags, D::Error>
    where
        D: Deserializer<'de>,
    {
        let _: () = <()>::deserialize(deserializer)?;
        Ok(CacheKeyFlags::empty())
    }
}

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "Attrs"))]
pub struct AttrsSerde<'a> {
    #[cfg_attr(feature = "serde", serde(with = "ColorOpt"))]
    pub color_opt: Option<Color>,
    #[cfg_attr(feature = "serde", serde(with = "FamilySerde"))]
    #[cfg_attr(feature = "serde", serde(borrow))]
    pub family: Family<'a>,
    #[cfg_attr(feature = "serde", serde(with = "StretchSerde"))]
    pub stretch: Stretch,
    #[cfg_attr(feature = "serde", serde(with = "StyleSerde"))]
    pub style: Style,
    #[cfg_attr(feature = "serde", serde(with = "WeightSerde"))]
    pub weight: Weight,
    pub metadata: usize,
    #[cfg_attr(feature = "serde", serde(with = "cache_key_flags"))]
    pub cache_key_flags: CacheKeyFlags,
}

impl<'a> From<Attrs<'a>> for AttrsSerde<'a> {
    fn from(value: Attrs<'a>) -> Self {
        AttrsSerde {
            color_opt: value.color_opt,
            family: value.family,
            stretch: value.stretch,
            style: value.style,
            weight: value.weight,
            metadata: value.metadata,
            cache_key_flags: value.cache_key_flags,
        }
    }
}

#[cfg(feature = "databake")]
impl<'a> Bake for AttrsSerde<'a> {
    fn bake(&self, ctx: &databake::CrateEnv) -> databake::TokenStream {
        let color = self.color_opt.map(|c| ColorSerde(c.0)).bake(ctx);
        let family: FamilySerde = self.family.into();
        let family = family.bake(ctx);
        let stretch: StretchSerde = self.stretch.into();
        let stretch = stretch.bake(ctx);
        let style: StyleSerde = self.style.into();
        let style = style.bake(ctx);
        let weight: WeightSerde = WeightSerde(self.weight.0);
        let weight = weight.bake(ctx);
        let metadata = self.metadata.bake(ctx);
        databake::quote! {
            cosmic_text::Attrs {
                color_opt: #color,
                family: #family,
                stretch: #stretch,
                style: #style,
                weight: #weight,
                metadata: #metadata,
                cache_key_flags: cosmic_text::CacheKeyFlags::empty()
            }
        }
    }
}

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "Option<Align>"))]
pub enum AlignSerde {
    Some(#[cfg_attr(feature = "serde", serde(with = "AlignRef"))] Align),
    None,
}

#[derive(Clone, Debug)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "serde", serde(remote = "Align"))]
#[cfg_attr(feature = "databake", derive(Bake))]
#[cfg_attr(feature = "databake", databake(path = cosmic_text::Align))]
pub enum AlignRef {
    Left,
    Right,
    Center,
    Justified,
    End,
}
